r"""
Construct 2 -- Parsing Made Fun

Homepage:
	https://github.com/construct/construct
    http://construct.readthedocs.org

Hands-on example:
    >>> from construct import *
    >>> s = Struct(
    ...     "a" / Byte,
    ...     "b" / Short,
    ... )
    >>> print s.parse(b"\x01\x02\x03")
    Container:
        a = 1
        b = 515
    >>> s.build(Container(a=1, b=0x0203))
    b"\x01\x02\x03"
"""

from construct.core import *
from construct.expr import *
from construct.debug import *
from construct.version import *
from construct import lib


#===============================================================================
# metadata
#===============================================================================
__author__ = "Arkadiusz Bulski <arek.bulski@gmail.com>, Tomer Filiba <tomerfiliba@gmail.com>, Corbin Simpson <MostAwesomeDude@gmail.com>"
__version__ = version_string

#===============================================================================
# exposed names
#===============================================================================
__all__ = [
    '__author__',
    '__version__',
    'abs_',
    'AdaptationError',
    'Adapter',
    'Aligned',
    'AlignedStruct',
    'Array',
    'Bit',
    'BitsInteger',
    'BitsSwapped',
    'BitStruct',
    'BitwisableString',
    'Bitwise',
    'Byte',
    'Bytes',
    'BytesInteger',
    'ByteSwapped',
    'Bytewise',
    'CancelParsing',
    'Check',
    'CheckError',
    'Checksum',
    'ChecksumError',
    'CipherError',
    'Compiled',
    'Compressed',
    'CompressedLZ4',
    'Computed',
    'Const',
    'ConstError',
    'Construct',
    'ConstructError',
    'Container',
    'CString',
    'Debugger',
    'Default',
    'Double',
    'EncryptedSym',
    'EncryptedSymAead',
    'Enum',
    'EnumInteger',
    'EnumIntegerString',
    'Error',
    'ExplicitError',
    'ExprAdapter',
    'ExprSymmetricAdapter',
    'ExprValidator',
    'Filter',
    'FixedSized',
    'Flag',
    'FlagsEnum',
    'FocusedSeq',
    'FormatField',
    'FormatFieldError',
    'FuncPath',
    'globalPrintFalseFlags',
    'globalPrintFullStrings',
    'GreedyBytes',
    'GreedyRange',
    'GreedyString',
    'Half',
    'Hex',
    'HexDump',
    'If',
    'IfThenElse',
    'Index',
    'IndexFieldError',
    'Indexing',
    'Int',
    'IntegerError',
    'Lazy',
    'LazyArray',
    'LazyBound',
    'LazyContainer',
    'LazyListContainer',
    'LazyStruct',
    'len_',
    'lib',
    'list_',
    'ListContainer',
    'Long',
    'Mapping',
    'MappingError',
    'max_',
    'min_',
    'NamedTuple',
    'NamedTupleError',
    'Nibble',
    'NoneOf',
    'NullStripped',
    'NullTerminated',
    'Numpy',
    'obj_',
    'Octet',
    'OffsettedEnd',
    'OneOf',
    'Optional',
    'Padded',
    'PaddedString',
    'Padding',
    'PaddingError',
    'PascalString',
    'Pass',
    'Path',
    'Path2',
    'Peek',
    'Pickled',
    'Pointer',
    'possiblestringencodings',
    'Prefixed',
    'PrefixedArray',
    'Probe',
    'ProcessRotateLeft',
    'ProcessXor',
    'RangeError',
    'RawCopy',
    'Rebuffered',
    'RebufferedBytesIO',
    'Rebuild',
    'release_date',
    'Renamed',
    'RepeatError',
    'RepeatUntil',
    'RestreamData',
    'Restreamed',
    'RestreamedBytesIO',
    'RotationError',
    'Seek',
    'Select',
    'SelectError',
    'Sequence',
    'setGlobalPrintFalseFlags',
    'setGlobalPrintFullStrings',
    'setGlobalPrintPrivateEntries',
    'Short',
    'Single',
    'SizeofError',
    'Slicing',
    'StopFieldError',
    'StopIf',
    'stream_iseof',
    'stream_read',
    'stream_read_entire',
    'stream_seek',
    'stream_size',
    'stream_tell',
    'stream_write',
    'StreamError',
    'StringEncoded',
    'StringError',
    'Struct',
    'Subconstruct',
    'sum_',
    'Switch',
    'SwitchError',
    'SymmetricAdapter',
    'Tell',
    'Terminated',
    'TerminatedError',
    'this',
    'Timestamp',
    'TimestampAdapter',
    'TimestampError',
    'Transformed',
    'Tunnel',
    'Union',
    'UnionError',
    'ValidationError',
    'Validator',
    'VarInt',
    'version',
    'version_string',
    'ZigZag',
]
__all__ += ["Int%s%s%s" % (n,us,bln) for n in (8,16,24,32,64) for us in "us" for bln in "bln"]
__all__ += ["Float%s%s" % (n,bln) for n in (16,32,64) for bln in "bln"]



from construct.lib.containers import *
from construct.lib.binary import *
from construct.lib.bitstream import *
from construct.lib.hex import *
from construct.lib.py3compat import *

__all__ = [
    'bits2bytes',
    'bits2integer',
    'byte2int',
    'bytes2bits',
    'bytes2integer',
    'bytes2integers',
    'bytes2str',
    'bytestringtype',
    'Container',
    'globalPrintFalseFlags',
    'globalPrintFullStrings',
    'HexDisplayedBytes',
    'HexDisplayedDict',
    'HexDisplayedInteger',
    'hexdump',
    'HexDumpDisplayedBytes',
    'HexDumpDisplayedDict',
    'hexlify',
    'hexundump',
    'int2byte',
    'integer2bits',
    'integer2bytes',
    'integers2bytes',
    'integertypes',
    'ListContainer',
    'ONWINDOWS',
    'PY',
    'PY2',
    'PY3',
    'PYPY',
    'RebufferedBytesIO',
    'reprstring',
    'RestreamedBytesIO',
    'setGlobalPrintFalseFlags',
    'setGlobalPrintFullStrings',
    'setGlobalPrintPrivateEntries',
    'str2bytes',
    'stringtypes',
    'swapbitsinbytes',
    'swapbytes',
    'swapbytesinbits',
    'trimstring',
    'unhexlify',
    'unicodestringtype',
]



from construct import *
from construct.lib import *
import binascii


def integer2bits(number, width, signed=False):
    r"""
    Converts an integer into its binary representation in a bit-string. Width is the amount of bits to generate. Each bit is represented as either \\x00 or \\x01. The most significant bit is first, big-endian. This is reverse to `bits2integer`.

    Examples:

        >>> integer2bits(19, 8)
        b'\x00\x00\x00\x01\x00\x00\x01\x01'
    """
    if not width >= 1:
        raise ValueError(f"width {width} must be positive")

    if signed:
        min = -(2 ** width // 2)
        max = 2 ** width // 2 - 1
    else:
        min = 0
        max = 2 ** width - 1
    if not min <= number <= max:
        raise ValueError(f"number {number} is out of range (min={min}, max={max})")

    if number < 0:
        number += 1 << width
    bits = bytearray(width)
    i = width - 1
    while number and i >= 0:
        bits[i] = number & 1
        number >>= 1
        i -= 1
    return bytes(bits)


def integer2bytes(number, width, signed=False):
    r"""
    Converts an integer into a byte-string. This is reverse to `bytes2integer`.

    Examples:

        >>> integer2bytes(19, 4)
        '\x00\x00\x00\x13'
    """
    # pypy does not check this in int.to_bytes, lazy fuckers
    if not width >= 1:
        raise ValueError(f"width {width} must be positive")

    try:
        return int.to_bytes(number, width, 'big', signed=signed)
    except OverflowError:
        raise ValueError(f"number {number} does not fit width {width}, signed {signed}")


def bits2integer(data, signed=False):
    r"""
    Converts a bit-string into an integer. Set signed to interpret the number as a 2-s complement signed integer. This is reverse to `integer2bits`.

    Examples:

        >>> bits2integer(b"\x01\x00\x00\x01\x01")
        19
    """
    if data == b"":
        raise ValueError("bit-string cannot be empty")

    number = 0
    for b in data:
        number = (number << 1) | b

    if signed and data[0]:
        bias = 1 << len(data)
        return number - bias
    else:
        return number


def bytes2integer(data, signed=False):
    r"""
    Converts a byte-string into an integer. This is reverse to `integer2bytes`.

    Examples:

        >>> bytes2integer(b'\x00\x00\x00\x13')
        19
    """
    if data == b"":
        raise ValueError("byte-string cannot be empty")

    return int.from_bytes(data, 'big', signed=signed)


BYTES2BITS_CACHE = {i:integer2bits(i,8) for i in range(256)}
def bytes2bits(data):
    r""" 
    Converts between bit-string and byte-string representations, both as bytes type.

    Example:

        >>> bytes2bits(b'ab')
        b"\x00\x01\x01\x00\x00\x00\x00\x01\x00\x01\x01\x00\x00\x00\x01\x00"
    """
    return b"".join(BYTES2BITS_CACHE[b] for b in data)


BITS2BYTES_CACHE = {bytes2bits(int2byte(i)):i for i in range(256)}
def bits2bytes(data):
    r""" 
    Converts between bit-string and byte-string representations, both as bytes type. Its length must be multiple of 8.

    Example:

        >>> bits2bytes(b"\x00\x01\x01\x00\x00\x00\x00\x01\x00\x01\x01\x00\x00\x00\x01\x00")
        b'ab'
    """
    if len(data) % 8 != 0:
        raise ValueError(f"data length {len(data)} must be a multiple of 8")
    return bytes(BITS2BYTES_CACHE[data[i:i+8]] for i in range(0,len(data),8))


def swapbytes(data):
    r"""
    Performs an endianness swap on byte-string.

    Example:

        >>> swapbytes(b'abcd')
        b'dcba'
    """
    return data[::-1]


def swapbytesinbits(data):
    r"""
    Performs an byte-swap within a bit-string. Its length must be multiple of 8.

    Example:

        >>> swapbytesinbits(b'0000000011111111')
        b'1111111100000000'
    """
    if len(data) % 8 != 0:
        raise ValueError(f"little-endianness is only defined if data length {len(data)} is multiple of 8")
    return b"".join(data[i:i+8] for i in reversed(range(0,len(data),8)))


SWAPBITSINBYTES_CACHE = {i:byte2int(bits2bytes(swapbytes(bytes2bits(int2byte(i))))) for i in range(256)}
def swapbitsinbytes(data):
    r"""
    Performs a bit-reversal on each byte within a byte-string.

    Example:

        >>> swapbitsinbytes(b"\xf0\x00")
        b"\x0f\x00"
    """
    return bytes(SWAPBITSINBYTES_CACHE[b] for b in data)


def hexlify(data):
    """Returns binascii.hexlify(data)."""
    return binascii.hexlify(data)


def unhexlify(data):
    """Returns binascii.unhexlify(data)."""
    return binascii.unhexlify(data)
